@@ -3,31 +3,34 @@
 # vim: set expandtab ts=4 sw=4:
 
 """
-Compute instantanous spectral metrics (Phase,Amplitude and Frequency) and
-compute frequency or time frequency spectra.
+Routines relating to frequency transforms and power-spectra.
 
-Routines:
+Frequency Transform Routines:
+  frequency_transform
+  quadrature_transform
+  phase_from_complex_signal
+  freq_from_phase
+  phase_from_freq
+  phase_angle
 
-frequency_stats
-quadrature_transform
-phase_from_complex_signal
-freq_from_phase
-phase_from_freq
-direct_quadrature
-phase_angle
-holospectrum
-hilberthuang
-hilberthuang_1d
-define_hist_bins
-define_hist_bins_from_data
+Power Spectra:
+  holospectrum
+  hilberthuang
+  hilberthuang_1d
+
+Power Spectra Helpers:
+  define_hist_bins
+  define_hist_bins_from_data
 
 """
 
 import logging
+import warnings
 import numpy as np
 from scipy import signal, sparse
 
-from . import utils
+from . import utils, cycles
+from .support import ensure_2d, ensure_equal_dims, ensure_vector
 
 # Housekeeping for logging
 logger = logging.getLogger(__name__)
@@ -35,21 +38,33 @@ logger = logging.getLogger(__name__)
 ##
 
 
-def frequency_stats(imf, sample_rate, method,
-                    smooth_phase=31):
-    """
-    Compute instantaneous phase, frequency and amplitude from a set of IMFs.
+def frequency_stats(*args, **kwargs):
+    """Legacy function for backwards compatibility."""
+    msg = "WARNING: 'emd.spectra.frequency_stats' is deprecated and " + \
+          "will be removed in a future version of EMD. Please change to use " + \
+          "'emd.spectra.frequency_transform' to remove this warning and " + \
+          "future-proof your code"
+
+    warnings.warn(msg)
+    logger.warning(msg)
+    return frequency_transform(*args, **kwargs)
+
+
+def frequency_transform(imf, sample_rate, method,
+                        smooth_phase=5):
+    """Compute instantaneous phase, frequency and amplitude from a set of IMFs.
+
     Several approaches are implemented from [1]_ and [2]_.
 
     Parameters
     ----------
     imf : ndarray
         Input array of IMFs.
-    sample_rate : scalar
+    sample_rate : float
         Sampling frequency of the signal in Hz
     method : {'hilbert','quad','direct_quad','nht'}
         The method for computing the frequency stats
-    smooth_phase : integer
+    smooth_phase : int
          Length of window when smoothing the unwrapped phase (Default value = 31)
 
     Returns
@@ -75,6 +90,8 @@ def frequency_stats(imf, sample_rate, method,
 
     """
     logger.info('STARTED: compute frequency stats')
+
+    imf = ensure_2d([imf], ['imf'], 'frequency_transform')
     logger.debug('computing on {0} samples over {1} imfs at sample rate {2}'.format(imf.shape[0],
                                                                                     imf.shape[1],
                                                                                     sample_rate))
@@ -108,6 +125,33 @@ def frequency_stats(imf, sample_rate, method,
         if orig_dim == 2:
             iamp = iamp[:, :, 0]
 
+    elif method == 'ctrl':
+        logger.info('Using Control Points - CURRENTLY BROKEN')
+
+        orig_dim = imf.ndim
+        if imf.ndim == 2:
+            imf = imf[:, :, None]
+
+        # Get phase from control points
+        iphase = np.zeros_like(imf)
+        for ii in range(imf.shape[1]):
+            for jj in range(imf.shape[2]):
+                good_cycles = cycles.get_cycle_inds_from_waveform(imf[:, ii, jj], cycle_start='asc')
+                ctrl = cycles.get_control_points(imf[:, ii, jj], good_cycles)
+                iphase[:, ii, jj] = phase_from_control_points(ctrl, good_cycles)
+                iphase[:, ii, jj] = np.unwrap(iphase[:, ii, jj])
+
+        # Estimate inst amplitudes with spline interpolation
+        iamp = np.zeros_like(imf)
+        for ii in range(imf.shape[1]):
+            for jj in range(imf.shape[2]):
+                iamp[:, ii, jj] = utils.interp_envelope(imf[:, ii, jj],
+                                                        mode='upper')
+
+        if orig_dim == 2:
+            iamp = iamp[:, :, 0]
+            iphase = iphase[:, :, 0]
+
     elif method == 'quad':
         logger.info('Using Quadrature transform')
 
@@ -142,9 +186,10 @@ def frequency_stats(imf, sample_rate, method,
         logger.error("Method '{0}' not recognised".format(method))
         raise ValueError("Method '{0}' not recognised\nPlease use one of 'hilbert','nht' or 'quad'".format(method))
 
-    # Compute unwrapped phase for frequency estimation
-    iphase = phase_from_complex_signal(
-        analytic_signal, smoothing=smooth_phase, ret_phase='unwrapped')
+    if method != 'ctrl':
+        # Compute unwrapped phase for frequency estimation
+        iphase = phase_from_complex_signal(
+            analytic_signal, smoothing=smooth_phase, ret_phase='unwrapped')
     ifreq = freq_from_phase(iphase, sample_rate)
 
     # Return wrapped phase
@@ -157,10 +202,11 @@ def frequency_stats(imf, sample_rate, method,
 
 
 def quadrature_transform(X):
-    """
-    Compute the quadrature transform on a set of time-series as defined in
-    equation 34 of [1]_. The return is a complex array with the input data as
-    the real part and the quadrature transform as the imaginary part.
+    """Compute the quadrature transform on a set of time-series.
+
+    This algorithm is defined in equation 34 of [1]_. The return is a complex
+    array with the input data as the real part and the quadrature transform as
+    the imaginary part.
 
     Parameters
     ----------
@@ -179,7 +225,6 @@ def quadrature_transform(X):
        1(2), 177–229. https://doi.org/10.1142/s1793536909000096
 
     """
-
     nX = utils.amplitude_normalise(X.copy(), clip=True)
 
     imagX = np.lib.scimath.sqrt(1 - np.power(nX, 2)).real
@@ -195,9 +240,10 @@ def quadrature_transform(X):
 
 def phase_from_complex_signal(complex_signal, smoothing=None,
                               ret_phase='wrapped', phase_jump='ascending'):
-    """
-    Compute the instantaneous phase from a complex signal obtained from either
-    the Hilbert Transform or by Direct Quadrature.
+    """Compute the instantaneous phase from a complex signal.
+
+    The complex input may be obtained from either the Hilbert Transform or by
+    Direct Quadrature.
 
     Parameters
     ----------
@@ -216,7 +262,6 @@ def phase_from_complex_signal(complex_signal, smoothing=None,
         Array of instantaneous phase values
 
     """
-
     # Compute unwrapped phase
     iphase = np.unwrap(np.angle(complex_signal), axis=0)
 
@@ -252,15 +297,13 @@ def phase_from_complex_signal(complex_signal, smoothing=None,
 
 
 def freq_from_phase(iphase, sample_rate):
-    """
-    Compute the instantaneous frequency from the differential of the
-    instantaneous phase.
+    """Compute the instantaneous frequency from the instantaneous phase.
 
     Parameters
     ----------
     iphase : ndarray
         Input array containing the unwrapped instantaneous phase time-course
-    sample_rate : scalar
+    sample_rate : float
         The sampling frequency of the data
 
     Returns
@@ -269,7 +312,6 @@ def freq_from_phase(iphase, sample_rate):
         Array containing the instantaneous frequencies
 
     """
-
     # Differential of instantaneous phase
     iphase = np.gradient(iphase, axis=0)
 
@@ -280,16 +322,15 @@ def freq_from_phase(iphase, sample_rate):
 
 
 def phase_from_freq(ifrequency, sample_rate, phase_start=-np.pi):
-    """
-    Compute the instantaneous phase of a signal from its instantaneous phase.
+    """Compute the instantaneous phase of a signal from its instantaneous frequency.
 
     Parameters
     ----------
     ifrequency : ndarray
         Input array containing the instantaneous frequencies of a signal
-    sample_rate : scalar
+    sample_rate : float
         The sampling frequency of the data
-    phase_start : scalar
+    phase_start : float
          Start value of the phase output (Default value = -np.pi)
 
     Returns
@@ -298,7 +339,6 @@ def phase_from_freq(ifrequency, sample_rate, phase_start=-np.pi):
         The instantaneous phase of the signal
 
     """
-
     iphase_diff = (ifrequency / sample_rate) * (2 * np.pi)
 
     iphase = phase_start + np.cumsum(iphase_diff, axis=0)
@@ -306,20 +346,42 @@ def phase_from_freq(ifrequency, sample_rate, phase_start=-np.pi):
     return iphase
 
 
+def phase_from_control_points(ctrl, cycles):
+    from scipy import interpolate as interp
+
+    cycles = ensure_vector([cycles],
+                           ['cycles'],
+                           'phase_from_control_points')
+
+    ip = np.zeros_like(cycles, dtype=float)
+    phase_y = np.array([0, np.pi / 2, np.pi, 3 * np.pi / 2, 2 * np.pi])
+
+    for jj in range(1, cycles.max() + 1):
+
+        if np.any(np.isnan(ctrl[jj-1, :])):
+            continue
+
+        f = interp.interp1d(ctrl[jj-1, :], phase_y, kind='linear')
+        ph = f(np.arange(0, ctrl[jj-1, -1] + 1))
+
+        ip[cycles == jj] = ph
+
+    return ip
+
+
 def direct_quadrature(fm):
-    """Section 3.2 of 'on instantaneous frequency'
-    Compute the quadrature transform on a set of time-series as defined in
-    equation 35 of [1].
+    """Compute the quadrature transform on a set of time-series.
+
+    This algorithm is  defined in equation 35 of [1].
+
+    Section 3.2 of 'on instantaneous frequency'
 
     THIS IS IN DEVELOPMENT
 
     Parameters
     ----------
-    fm :
-
-
-    Returns
-    -------
+    fm : ndarray
+        Input signal containing a frequency-modulated signal
 
     References
     ----------
@@ -340,9 +402,9 @@ def direct_quadrature(fm):
 
 
 def phase_angle(fm):
-    """
-    Compute the quadrature transform on a set of time-series as defined in
-    equation 35 of [1]_.
+    """Compute the phase angle of a set of time-series.
+
+    This algorithm is defined in equation 35 of [1]_.
 
     THIS IS IN DEVELOPMENT
 
@@ -362,9 +424,7 @@ def phase_angle(fm):
        K. (2009). On Instantaneous Frequency. Advances in Adaptive Data Analysis,
        1(2), 177–229. https://doi.org/10.1142/s1793536909000096
 
-
     """
-
     return np.arctan(fm / np.lib.scimath.sqrt(1 - np.power(fm, 2)))
 
 # Time-frequency spectra
@@ -372,8 +432,9 @@ def phase_angle(fm):
 
 def holospectrum(infr, infr2, inam2, freq_edges, freq_edges2, mode='energy',
                  squash_time='sum'):
-    """
-    Compute the Holospectrum from the first and second layer frequecy
+    """Compute a Holospectrum.
+
+    Holospectra are computed from the first and second layer frequecy
     statistics of a dataset. The Holospectrum represents the energy of a signal
     across time, carrier frequency and amplitude-modulation frequency [1]_.
 
@@ -417,15 +478,21 @@ def holospectrum(infr, infr2, inam2, freq_edges, freq_edges2, mode='energy',
 
     """
     logger.info('STARTED: compute Holospectrum')
+
+    out = ensure_2d((infr, infr2, inam2), ('infr', 'infr2', 'inam2'), 'holospectrum')
+    infr, infr2, inam2 = out
+    ensure_equal_dims((infr, infr2, inam2), ('infr', 'infr2', 'inam2'), 'holospectrum', dim=0)
+    ensure_equal_dims((infr, infr2, inam2), ('infr', 'infr2', 'inam2'), 'holospectrum', dim=1)
+
     logger.debug('computing on {0} samples over {1} first-level \
                   IMFs and {2} second level IMFs'.format(infr2.shape[0],
                                                          infr2.shape[1],
                                                          infr2.shape[2]))
     logger.debug('First level freq bins: {0} to {1} in {2} steps'.format(freq_edges[0],
-                                                                         freq_edges[1],
+                                                                         freq_edges[-1],
                                                                          len(freq_edges)))
     logger.debug('Second level freq bins: {0} to {1} in {2} steps'.format(freq_edges2[0],
-                                                                          freq_edges2[1],
+                                                                          freq_edges2[-1],
                                                                           len(freq_edges2)))
 
     if mode == 'energy':
@@ -476,10 +543,11 @@ def holospectrum(infr, infr2, inam2, freq_edges, freq_edges2, mode='energy',
 
 
 def hilberthuang(infr, inam, freq_edges, mode='energy', return_sparse=False):
-    """
-    Compute the Hilbert-Huang transform from the instataneous frequency
-    statistics of a dataset. The Hilbert-Huang Transform represents the energy
-    of a signal across time and frequency [1]_.
+    """Compute a Hilbert-Huang transform.
+
+    The Hilbert-Huang transform instataneous frequency and instantaneous
+    amplitude time-series and represents the energy of a signal across time and
+    frequency [1]_.
 
     Parameters
     ----------
@@ -516,19 +584,22 @@ def hilberthuang(infr, inam, freq_edges, mode='energy', return_sparse=False):
        https://doi.org/10.1098/rspa.1998.0193
 
     """
+    infr, inam = ensure_2d([infr, inam], ['infr', 'inam'], 'hilberthuang')
+    ensure_equal_dims((infr, inam), ('infr', 'inam'), 'hilberthuang')
 
     logger.info('STARTED: compute Hilbert-Huang Transform')
     logger.debug('computing on {0} samples over {1} IMFs '.format(infr.shape[0],
                                                                   infr.shape[1]))
     logger.debug('Freq bins: {0} to {1} in {2} steps'.format(freq_edges[0],
-                                                             freq_edges[1],
+                                                             freq_edges[-1],
                                                              len(freq_edges)))
 
     if mode == 'energy':
         inam = inam**2
 
     # Create sparse co-ordinates
-    yinds = np.digitize(infr, freq_edges)
+    yinds = np.digitize(infr, freq_edges) - 1
+    yinds[yinds < 0] = 0
     xinds = np.tile(np.arange(yinds.shape[0]), (yinds.shape[1], 1)).T
 
     coo_data = (inam.reshape(-1), (yinds.reshape(-1), xinds.reshape(-1)))
@@ -548,10 +619,11 @@ def hilberthuang(infr, inam, freq_edges, mode='energy', return_sparse=False):
 
 
 def hilberthuang_1d(infr, inam, freq_edges, mode='energy'):
-    """
-    Compute the Hilbert-Huang transform from the instataneous frequency
-    statistics of a dataset. The 1D Hilbert-Huang Transform represents the
-    energy in a signal across frequencies and IMFs [1]_.
+    """Compute the Hilbert-Huang transform with no time-dimension.
+
+    The Hilbert-Huang transform instataneous frequency and instantaneous
+    amplitude time-series and represents the energy of a signal across
+    frequency [1]_.
 
     Parameters
     ----------
@@ -579,7 +651,6 @@ def hilberthuang_1d(infr, inam, freq_edges, mode='energy'):
        https://doi.org/10.1098/rspa.1998.0193
 
     """
-
     specs = np.zeros((len(freq_edges) - 1, infr.shape[1]))
 
     # Remove values outside the bin range
@@ -601,16 +672,15 @@ def hilberthuang_1d(infr, inam, freq_edges, mode='energy'):
 
 
 def define_hist_bins(data_min, data_max, nbins, scale='linear'):
-    """
-    Define the bin edges and centre values for use in a histogram
+    """Define the bin edges and centre values for use in a histogram.
 
     Parameters
     ----------
-    data_min : scalar
+    data_min : float
         Value for minimum edge
-    data_max : scalar
+    data_max : float
         Value for maximum edge
-    nbins : integer
+    nbins : int
         Number of bins to create
     scale : {'linear','log'}
          Flag indicating whether to use a linear or log spacing between bins (Default value = 'linear')
@@ -623,6 +693,7 @@ def define_hist_bins(data_min, data_max, nbins, scale='linear'):
         1D array of bin centres
 
     Notes
+    -----
     >> edges,centres = emd.spectra.define_hist_bins( 1, 5, 3 )
     >> print(edges)
     [1. 2. 3. 4. 5.]
@@ -630,7 +701,6 @@ def define_hist_bins(data_min, data_max, nbins, scale='linear'):
     [1.5 2.5 3.5 4.5]
 
     """
-
     if scale == 'log':
         p = np.log([data_min, data_max])
         edges = np.linspace(p[0], p[1], nbins + 1)
@@ -647,9 +717,9 @@ def define_hist_bins(data_min, data_max, nbins, scale='linear'):
 
 
 def define_hist_bins_from_data(X, nbins=None, mode='sqrt', scale='linear'):
-    """Find the bin edges and centre frequencies for use in a histogram
+    """Find the bin edges and centre frequencies for use in a histogram.
 
-    if nbins is defined, mode is ignored
+    If nbins is defined, mode is ignored
 
     Parameters
     ----------
@@ -670,7 +740,6 @@ def define_hist_bins_from_data(X, nbins=None, mode='sqrt', scale='linear'):
         1D array of bin centres
 
     """
-
     data_min = X.min()
     data_max = X.max()
 
