@@ -6,24 +6,22 @@
 Support functions.
 
 Routines:
+  amplitude_normalise
+  abreu2010
+  est_orthogonality
+  find_extrema_locked_epochs
+  apply_epochs
+  wrap_phase
+  ar_simulate
 
-amplitude_normalise
-get_padded_extrema
-interp_envelope
-find_extrema
-zero_crossing_count
-abreu2010
-est_orthogonality
-find_extrema_locked_epochs
-apply_epochs
-wrap_phase
 
 """
 
 import logging
 import numpy as np
+from scipy import signal
 
-from .sift import interp_envelope
+from .sift import interp_envelope, get_padded_extrema
 
 # Housekeeping for logging
 logger = logging.getLogger(__name__)
@@ -31,17 +29,17 @@ logger = logging.getLogger(__name__)
 
 def amplitude_normalise(X, thresh=1e-10, clip=False, interp_method='pchip',
                         max_iters=3):
-    """
-    Normalise the amplitude envelope of an IMF to be 1. Multiple runs of
-    normalisation are carried out until the desired threshold is reached.
+    """Normalise the amplitude envelope of an IMF to be 1.
 
-    This uses the method described as part of the AM-FM transform [1]_
+    Multiple runs of normalisation are carried out until the desired threshold
+    is reached. This uses the method described as part of the AM-FM transform
+    [1]_
 
     Parameters
     ----------
     X : ndarray
         Input array of IMFs to be normalised
-    thresh : scalar
+    thresh : float
          Threshold for stopping normalisation (Default value = 1e-10)
     clip : bool
          Whether to clip the output between -1 and 1 (Default value = False)
@@ -124,20 +122,19 @@ def amplitude_normalise(X, thresh=1e-10, clip=False, interp_method='pchip',
 
 
 def abreu2010(f, nonlin_deg, nonlin_phi, sample_rate, seconds):
-    """
-    Simulate a non-linear waveform using equation 9 in [1]_.
+    """Simulate a non-linear waveform using equation 9 in [1]_.
 
     Parameters
     ----------
-    f : scalar
+    f : float
         Fundamental frequency of generated signal
-    nonlin_deg : scalar
+    nonlin_deg : float
         Degree of non-linearity in generated signal
-    nonlin_phi : scalar
+    nonlin_phi : float
         Skew in non-linearity of generated signal
-    sample_rate : scalar
+    sample_rate : float
         The sampling frequency of the generated signal
-    seconds : scalar
+    seconds : float
         The number of seconds of data to generate
 
     Returns
@@ -152,7 +149,6 @@ def abreu2010(f, nonlin_deg, nonlin_phi, sample_rate, seconds):
        57(7), 656–667. https://doi.org/10.1016/j.coastaleng.2010.02.005
 
     """
-
     time_vect = np.linspace(0, seconds, int(seconds * sample_rate))
 
     factor = np.sqrt(1 - nonlin_deg**2)
@@ -165,8 +161,7 @@ def abreu2010(f, nonlin_deg, nonlin_phi, sample_rate, seconds):
 
 
 def est_orthogonality(imf):
-    """
-    Compute the index of orthogonality as described in equation 6.5 of [1]_.
+    """Compute the index of orthogonality as described in equation 6.5 of [1]_.
 
     Parameters
     ----------
@@ -188,7 +183,6 @@ def est_orthogonality(imf):
        https://doi.org/10.1098/rspa.1998.0193
 
     """
-
     ortho = np.ones((imf.shape[1], imf.shape[1])) * np.nan
 
     for ii in range(imf.shape[1]):
@@ -199,19 +193,18 @@ def est_orthogonality(imf):
     return ortho
 
 
-def find_extrema_locked_epochs(X, winsize, lock_to='max', percentile=None):
-    """
-    Helper function for defining epochs around peaks or troughs within the data
+def find_extrema_locked_epochs(X, winsize, lock_to='peaks', percentile=None):
+    """Define epochs around peaks or troughs within the data.
 
     Parameters
     ----------
     X : ndarray
         Input time-series
-    winsize : integer
+    winsize : int
         Width of window to extract around each extrema
     lock_to : {'max','min'}
          Flag to select peak or trough locking (Default value = 'max')
-    percentile : scalar
+    percentile : float
          Optional flag to selection only the upper percentile of extrema by
          magnitude (Default value = None)
 
@@ -221,38 +214,34 @@ def find_extrema_locked_epochs(X, winsize, lock_to='max', percentile=None):
         Array of start and end indices for epochs around extrema.
 
     """
-    if lock_to not in ['max', 'min']:
+    if lock_to not in ['peaks', 'troughs', 'combined']:
         raise ValueError("Invalid lock_to value")
 
-    from .sift import find_extrema
-    if lock_to == 'max':
-        locs, pks = find_extrema(X, ret_min=False)
-    else:
-        locs, pks = find_extrema(X, ret_min=True)
+    locs, pks = get_padded_extrema(X, pad_width=0, mode=lock_to)
 
     if percentile is not None:
-        thresh = np.percentile(pks[:, 0], percentile)
-        locs = locs[pks[:, 0] > thresh]
+        thresh = np.percentile(pks, percentile)
+        locs = locs[pks > thresh]
         pks = pks[pks > thresh]
 
     winstep = int(winsize / 2)
 
+    # Get all trials
     trls = np.r_[np.atleast_2d(locs - winstep), np.atleast_2d(locs + winstep)].T
 
     # Reject trials which start before 0
     inds = trls[:, 0] < 0
-    trls = trls[inds is False, :]
+    trls = trls[inds == False, :]  # noqa: E712
 
     # Reject trials which end after X.shape[0]
     inds = trls[:, 1] > X.shape[0]
-    trls = trls[inds is False, :]
+    trls = trls[inds == False, :]  # noqa: E712
 
     return trls
 
 
 def apply_epochs(X, trls):
-    """
-    Apply a set of epochs to a continuous dataset
+    """Apply a set of epochs to a continuous dataset.
 
     Parameters
     ----------
@@ -268,7 +257,6 @@ def apply_epochs(X, trls):
         Epoched time-series
 
     """
-
     Y = np.zeros((trls[0, 1] - trls[0, 0], X.shape[1], trls.shape[0]))
     for ii in np.arange(trls.shape[0]):
 
@@ -285,7 +273,7 @@ def wrap_phase(IP, ncycles=1, mode='2pi'):
     ----------
     IP : ndarray
         Input array of unwrapped phase values
-    ncycles : integer
+    ncycles : int
          Number of cycles per wrap (Default value = 1)
     mode : {'2pi','-pi2pi'}
          Flag to indicate the values to wrap phase within (Default value = '2pi')
@@ -305,3 +293,55 @@ def wrap_phase(IP, ncycles=1, mode='2pi'):
         phases = (IP + (np.pi * ncycles)) % (ncycles * 2 * np.pi) - (np.pi * ncycles)
 
     return phases
+
+
+def ar_simulate(freq, sample_rate, seconds, r=.95, noise_std=None, random_seed=None):
+    """Create a simulated oscillation using an autoregressive filter.
+
+    A simple filter is defined by direct pole placement and applied to white
+    noise to generate a random signal with a defined oscillatory peak frequency
+    that exhibits random variability frequency, amplitude and waveform.
+
+    Parameters
+    ----------
+    freq : float
+        Peak resonant frequency of the simulated filter.
+    sample_rate : float
+        Sampling frequency for the simulation
+    seconds : float
+        Number of seconds of data to simulate
+    r : float (0 < r < 1)
+        Pole magnitude of simulated autoregressive resonance.
+    noise_std : float
+        Scaling of optional noise to add to simulation. Scaling is relative to
+        standard-deviation of the simulated data.
+    random_seed : int
+        Optional random seed generation
+
+    Returns
+    -------
+    ndarray
+        A simulated time course.
+
+    """
+    if random_seed is not None:
+        np.random.seed(random_seed)
+
+    if freq > 0:
+        freq_rads = (2 * np.pi * freq) / sample_rate
+        a1 = np.array([1, -2*r*np.cos(freq_rads), (r**2)])
+    else:
+        a1 = np.poly(r)
+
+    num_samples = int(sample_rate * seconds)
+
+    x = signal.filtfilt(1, a1, np.random.randn(1, num_samples)).T
+
+    if noise_std is not None:
+        noise = np.std(x)*noise_std*np.random.randn(1, num_samples).T
+        x = x + noise
+
+    if random_seed is not None:
+        np.random.seed()  # restore defaults
+
+    return x
